//////////////////////////////////////////////////////////////////////////////
//		 Arduino Intervall-Timer mit Bluetooth-Modul
//			PID Temperaturregelung
//
//       17.08.2024	                             Erwin Lotter
//////////////////////////////////////////////////////////////////////////////

#include "IntervallTimer.h"


int8_t bmp280_avl = -3;
int8_t aht20_avl = -20;
//inline void setupAHT20() __attribute__((always_inline));

/*
#if defined(AHT20_AVAILABLE)
#include <BMP280_DEV.h>		  // Include the BMP280_DEV.h library, https://registry.platformio.org/libraries/martinl1/BMP280_DEV/examples/BMP280_I2C_Normal/BMP280_I2C_Normal.ino
BMP280_DEV bmp280;		  // Instantiate (create) a BMP280_DEV object and set-up for I2C operation (address 0x77)
				  // braucht 3370 Bytes  bzw. 996 Bytes zusätzlich zur AHT20-Lib
#endif

bool IntervallTimer::ReadAmbientTemperatureX10(int& temp)
{
#if defined(AHT20_AVAILABLE)
  if (bmp280_avl < 0) {
    if (bmp280.begin()) {
      bmp280_avl = 1;
      //bmp280.setPresOversampling(OVERSAMPLING_X4);    // Set the pressure oversampling to X4
      //bmp280.setTempOversampling(OVERSAMPLING_X1);    // Set the temperature oversampling to X1
      //bmp280.setIIRFilter(IIR_FILTER_4);              // Set the IIR filter to setting 4
      bmp280.setTimeStandby(TIME_STANDBY_1000MS);     // Set the standby time to 2 seconds
      bmp280.startNormalConversion();                 // Continue BMP280 continuous conversion in NORMAL_MODE  
    }
    else {
      bmp280_avl++;
    }
  }
  if (bmp280_avl > 0) {
    //if (bmp280.getMeasurements(temperature, pressure, altitude))    // Check if the measurement is complete
     //temp = int(temperature * 10) + TuCorr;
    return true;
  }
#endif
  return false;
}

//bool IntervallTimer::ReadAmbientHumidity(float& humidity) {return false;}
//*/


/*  // ********** Adafruit ***********
#if defined(AHT20_AVAILABLE)  
#include <Adafruit_AHTX0.h>
Adafruit_AHTX0 aht20;		// braucht 5126 Bytes

#include <Adafruit_BMP280.h>    // Temperatur von BMP280:
Adafruit_BMP280 bmp280;		// braucht 8238 Bytes

// Adafruit_AHTX0 und Adafruit_AHTX0 brauchen zusammen 9790 Bytes
#endif
// 
bool IntervallTimer::ReadAmbientTemperatureX10(int& temp)
{
#if defined(AHT20_AVAILABLE)   // code from github Adafruit_BMP280_Library
  if (bmp280_avl < 0) {
    if (bmp280.begin()) {
      bmp280_avl = 1;
      bmp280.setSampling(Adafruit_BMP280::MODE_NORMAL,     // Operating Mode: alle xx msec
		      Adafruit_BMP280::SAMPLING_NONE,     // no Temp. oversampling
		      Adafruit_BMP280::SAMPLING_NONE,    // no Pressure oversampling
		      Adafruit_BMP280::FILTER_OFF,      // Filtering.
		      Adafruit_BMP280::STANDBY_MS_1000); // Standby time.
    }
    else {
      bmp280_avl++;
    }
  }
  if (bmp280_avl > 0) {
    temp = int(bmp280.readTemperature() * 10) + TuCorr;
    return true;
  }
#endif
  return false;
}

inline void setupAHT20() 
{
#if defined(AHT20_AVAILABLE)
  if (aht20_avl < 0) {
    if (aht20.begin()) {
      aht20_avl = 1;
    }
    else {
      aht20_avl++;
    }
  }
#endif
}

bool IntervallTimer::ReadAmbientHumidity(float& humidity)
{
#if defined(AHT20_AVAILABLE)
  setupAHT20();
  if (aht20_avl > 0) {
    //humidity = aht20.getHumidity();
    return true;
  }
#endif
  return false;
}
*/

//*
#if defined(AHT20_AVAILABLE)
#include <AHT20.h>
AHT20 aht20;		// braucht 3306 Bytes
#endif

inline void setupAHT20() 
{
#if defined(AHT20_AVAILABLE)
  if (aht20_avl < 0) {
    if (aht20.begin()) {
      aht20_avl = 1;
    }
    else {
      aht20_avl++;
    }
  }
#endif
}

bool IntervallTimer::ReadAmbientTemperatureX10(int& temp)
{
#if defined(AHT20_AVAILABLE)
  setupAHT20();
  if (aht20_avl > 0) {
    temp = int(aht20.getTemperature() * 10) + TuCorr;
    return true;
  }
#endif
  temp = temp;
  return false;
}

bool IntervallTimer::ReadAmbientHumidity(float& humidity)
{
#if defined(AHT20_AVAILABLE)
  setupAHT20();
  if (aht20_avl > 0) {
    humidity = aht20.getHumidity();
    return true;
  }
#endif
  humidity = humidity;
  return false;
}
//*/


#ifdef T_SENSOR_AVAILABLE
int IntervallTimer::ReadTemperatureX10()
{
  return (int)ReadTemperature_times_10() + tempCorr;
}
#endif


//
// Set cooling mode 1 (lowest) to 5 (highest) or switch cooling off (mode 0)
//  according to commands "0K" .. "5K". Initial mode (= 3) is applied for mode -1.
// 
// Returns the calculated COOLER_PWM output which should be in the range 0 - 255
//  or -1 if the cooling was switched off
//  or -2 if cooling is not available..
//
inline int ApplyCoolingMode(int mode)
{
#ifdef COOLER_POWER
  digitalWrite(COOLER_POWER, mode==0 ? !COOLER_POWER_ON : COOLER_POWER_ON);
# ifndef COOLER_PWM
  return mode != 0;
# endif
#endif
#ifdef COOLER_PWM
  if (mode == 0) {
    #ifdef COOLER_POWER
    return -1;		    // power switched off: leave mode unchanged
    #endif
    mode++;		    // no power switch: mode = min
  }
  if (mode == -1) {
    mode = 3;
  }
  int pwmout = CoolerPWMLevels[
    #if COOLER_PWM_REVERSE    // The cooling versus PWM signal curve has a negative slope:
      4 -		//  0..4 --> 4..0
    #endif
      (mode-1)];
  analogWrite(COOLER_PWM, pwmout);
  return pwmout;
#else
  mode = mode;
  return -2;
#endif 
}


//
// Set window heating mode 1 (lowest) to 5 (highest) or switch heating off (mode 0)
//  according to commands "20K" .. "25K". Initial mode (= 5) is applied for mode==-1.
// 
// Returns the calculated HEATER_PWM output which should be in the range 0 - 255
//  or -1 if the heating was switched off
//  or -2 if not available..
//
inline int ApplyHeatingMode(int mode)
{
#if defined(HEATER_POWER) && !defined(HEATER_SOFT_PWM)
  digitalWrite(HEATER_POWER, mode==0 ? !HEATER_POWER_ON : HEATER_POWER_ON);
# ifndef HEATER_PWM
  return mode != 0;
# endif
#endif
#if defined(HEATER_PWM) || defined(HEATER_SOFT_PWM)
  if (mode == 0) {
    #ifdef HEATER_POWER
    return -1;		    // power switched off: leave mode unchanged
    #endif
    mode++;		    // no power switch: mode = min
  }
  if (mode == -1) {
    mode = 5;
  }
  // PWM settings for "21K" to "25K" (in reverse order if reverse PWM signal)
  int pwmout = HeaterPWMLevels[
    #if defined(HEATER_PWM) && HEATER_PWM_REVERSE    // The heating versus PWM signal curve has a negative slope:
      4 -		//  0..4 --> 4..0
    #endif
      (mode-1)];
  #if defined(HEATER_PWM)		      // Soft-PWM will be actuated later
  analogWrite(HEATER_PWM, pwmout);
  #endif
  return pwmout;
#else
  mode = mode;
  return -2;
#endif 
}



#if defined(COOLER_PWM) || defined(COOLER_POWER)
//
// Set cooling mode by calling ApplyCoolingMode() and write result
//  into coolingLevel.
// 
void IntervallTimer::SetCoolingMode(int mode)
{
  coolingLevel = ApplyCoolingMode(mode);
}
#endif

#if defined(HEATER_PWM) || defined(HEATER_POWER)
//
// Set heating mode by calling ApplyHeatingMode() and write result
//  into heatingLevel.
// 
void IntervallTimer::SetHeatingMode(int mode)
{
  heatingLevel = ApplyHeatingMode(mode);
}
#endif


#define PID_Cycle 1000	  // delay between PID actuations in milliseconds - don't change!

//
// Update cooling PID and heating soft-pwm
// 
// Return action on cooling PID:
//  1: just updated 
//  0: not yet updated
// -1: not active
//
int IntervallTimer::UpdatePID()
{
#if defined(HEATER_SOFT_PWM)
  //
  // Divide time in ~ one-second slots. 
  // Divide each slot in 256 slices which get high or low depending on 
  //  heatingLevel setting.
  // millis() >> 10 --> 1 slot
  // millis() >> (10-8) --> 1 slice
  // (millis() >> (10-8)) & 0x00FF --> slice mask
  //
  bool on = (int((millis() >> (10 - 8))) & 0x00FF) < heatingLevel; // switch ON as long as slice has not reached level
  digitalWrite(HEATER_POWER, on ? HEATER_POWER_ON : !HEATER_POWER_ON);
#endif
#if !defined(COOLER_PWM)
  return -1;
#else
  if (IsCalibPID()) {    // PID calibration running?
    return UpdatePIDCalib(false);
  }
  if (PID_delay < 0) {
    return -1;		    // not startet
  }
  if (PID_delay > 0) {	    // still waitung for start?
    if (timeReached(PID_prevTime)) {
      InitPID(ReadTemperatureX10());
    }
    else {
      return -1;	    // still waiting
    }
  }
  //
  // PID Controller is running:
  //
  uint32_t time = millis();
  int32_t elapsedTime = time - PID_prevTime;
  if (elapsedTime < PID_Cycle) {	  // wait a second until next cycle
    return 0;
  }
  PID_prevTime = time;

  int error =
  #if COOLER_PWM_REVERSE  // the cooling versus PWM output signal has a
    -			  //  negative slope (i.e. higher output = less cooling)?
  #endif
    (PID_SetTemp - ReadTemperatureX10());

  #define kp  KpKiKd[0]
  #define ki  KpKiKd[1]
  #define kd  KpKiKd[2]
  float PID_p = kp * error;

  PID_i = PID_i + (ki * error); // *(elapsedTime/PID_Cycle)    // assume 1 second

  float PID_d = kd * (error - PID_error); // /(elapsedTime/PID_Cycle)    // assume 1 second

  int newCoolingLevel = (int)(PID_p + PID_i + PID_d);
  if (debug) {
    bluePrintTdegC(ReadTemperatureX10());
    bluePrint_P(PSTR(", p="));
    bluePrint((int)PID_p);
    bluePrint_P(PSTR(" + i="));
    bluePrint((int)PID_i);
    bluePrint_P(PSTR(" + d="));
    bluePrint((int)PID_d);
    bluePrint_P(PSTR(" -> "));
    bluePrintln(newCoolingLevel);
  }
# if defined(COOLER_POWER)
  bool coff = false;
# endif
  if(newCoolingLevel < 0) {    // PWM range 0 - 255:
    #if !(defined(COOLER_POWER) && COOLER_PWM_REVERSE)  // the cooling versus PWM output signal has a negative slope:
    if (newCoolingLevel < -15) {
      coff = true;	    // switch cooling off
    }
    #endif
    newCoolingLevel = 0;
  }
  if(newCoolingLevel > 255) {
    #if defined(COOLER_POWER) && COOLER_PWM_REVERSE  // the cooling versus PWM output signal has a negative slope:
    if (newCoolingLevel > 270) {
      coff = true;	    // switch cooling off
    }
    #endif
    newCoolingLevel = 255;  
  }
  SetCoolingLevel(newCoolingLevel);
# if defined(COOLER_POWER)
  if (coff != PID_coolingOff) {	    // switch cooling off if lowest level ist still too high
    SetCoolingMode(coff ? 0 : 1);   //  or back to lowest level
    PID_coolingOff = coff;
  }
# endif
  PID_error = error;     // previous error --> next loop.
  return 1;
#endif
}


#ifdef COOLER_PWM
//////////////////////////////////////////////////////////////////////////
// Continue PID temperature control after given delay
//
void IntervallTimer::StartPID(int afterSeconds)
{
  StopPID(); 
  if (!IsCooling()) {		  // activate std cooling, if cooling is off
    SetCoolingMode(-1);
  }
  PID_delay = afterSeconds+1;		  // status: PID
  PID_prevTime = millis() + afterSeconds * 1000L; // set delay
}


//
// Continue PID temperature control for given destination temperture.
//
void IntervallTimer::StartPIDT(float temp)
{
  StopPID(); 
  if (!IsCooling()) {		  // activate std cooling, if cooling is off
    SetCoolingMode(-1);
  }
  InitPID((int)(temp + 0.5));
}


//
// Abort any temperature PID action w/o notification
//
void IntervallTimer::AbortPID()
{
  PID_delay = -1;
}


//
// Stop temperature PID
//
void IntervallTimer::StopPID()
{
  if (IsCalibPID()) {
    //UpdatePIDCalib(true);
  }
  else {
    if (PID_delay >= 0) {
      EnterNewLineLn(PSTR("Temperaturregelung abgebrochen"));
      if (PID_coolingOff) {
        SetCoolingMode(1);		  // re-activate (minimal) cooling if switched off by PID control
      }
    }
    AbortPID();
  }
}

// PID controller in action atm?
bool IntervallTimer::IsPIDActive() const 
{
  return PID_delay == 0; 
}

// PID inactive and not waiting and not calibrating?
bool IntervallTimer::IsPIDOff() const 
{
  return PID_delay == -1; 
}

 
bool IntervallTimer::IsCalibPID() const 
{
  return PID_delay == -2; 
}

void IntervallTimer::DeletePID() 
{
  if (IsCalibPID()) {
    AbortPID();
  }
  delete PIDCal; 
  PIDCal = 0;
}


//
// Init PID settings
//
void IntervallTimer::InitPID(int tempX10)
{
  PID_coolingOff = false;
  PID_delay = 0;	    // srart now
  PID_SetTemp = tempX10;
  PID_i = coolingLevel;	    // equilibrium cooling level
  PID_error = 0;
  PID_prevTime = millis() - PID_Cycle; // sofort anfangen
  bluePrint_P(PSTR("Temperaturregelung gestartet, Ziel-Temperatur = "));
  bluePrintTdegC(tempX10);
  bluePrintln();
}


#define TEST_PIDCALIB 0
#if TEST_PIDCALIB
  // Testdaten für die PID-Kalibrierung
  // T in 0.1°C im Abstand von 60s
  const int testT[90] = {
    287, 285, 281, 275, 272, 268, 265, 262, 260, 257, 254, 252, 249, 247, 245, 
    243, 241, 239, 238, 236, 234, 233, 232, 230, 229, 228, 226, 225, 224, 
    223, 222, 221, 220, 220, 219, 218, 217, 216, 216, 215, 215, 214, 213, 
    213, 212, 212, 212, 211, 211, 210, 210, 210, 209, 209, 209, 208, 208, 
    207, 206, 205, 204, 204, 204, 203, 203, 203, 203, 203, 203, 203, 203, 
    203, 203, 203, 202, 202, 202, 202, 202, 202, 202, 202, 202, 202, 202, 
    202, 202, 202, 202, 202, 
  };
  int testI;	    // index in testT
  uint32_t nexttestmilli;
# define test_dt 600	    // msec between test samples (real value = 60000)
#endif

//
// Continue PID calibration with given output step
//
bool IntervallTimer::CalibPID(int step)
{
#if !TEST_PIDCALIB
  if (step < 10) {
    bluePrintln_P(PSTR("Kühlungs-Delta zu klein für PID-Kalibrierung!"));
    return false; 
  }
  if (!IsCooling()) {
    SetCoolingMode(-1);
  }
  int newCoolingLevel;
  bool badstep;
  #if COOLER_PWM_REVERSE
    newCoolingLevel = coolingLevel - step;  // the cooling versus PWM output signal has a negative slope 
    badstep = newCoolingLevel < 0;	    //  (i.e. higher output = less cooling): increase cooling by negative step
  #else
    newCoolingLevel = coolingLevel + step;
    badstep = newCoolingLevel > 255;
  #endif
  if (badstep) {
    bluePrintln_P(PSTR("Kühlungs-Delta zu groß für PID-Kalibrierung!"));
    return false; 
  }
#endif
  if (!PIDCal) {
    PIDCal = new StepResponse<100, RSCAL>;
    if (!PIDCal) {
      bluePrintln_P(PSTR("Nicht genug Speicher für die PID-Kalibrierung!"));
      return false; 
    }
  }
#if TEST_PIDCALIB
  testI = 0;
  nexttestmilli = millis() + test_dt;	  // supply reading every test_dt msec
  PIDCal->Init(1, -100);
  step = step;
  bluePrint_P(PSTR("PID-Kalibrierungstest gestartet"));
#else
  // The sign of the measure T will be inverted to get a increasing response.
  // To compensate this, step needs to be negated, too.
  // (Having applied a negative step to the coolingLevel in case 
  //  of COOLER_PWM_REVERSE != 0 is not taken into account, since this will
  //  be done automatically during temperature control, too.)
  //
  PIDCal->Init(initialResponseDelta_xS, -step);
  bluePrint_P(PSTR("PID-Kalibrierung gestartet, PWM="));
  bluePrint(coolingLevel);
  bluePrint_P(PSTR(" -> "));
  SetCoolingLevel(newCoolingLevel);
  bluePrintln(newCoolingLevel);
#endif
  PID_calCnt = PIDCal->GetCount();
  PID_delay = -2;	  // Flag Calib
  return true;
}


//
// Read again data for PID calibration, finish if stop==true.
//
int IntervallTimer::UpdatePIDCalib(bool stop)
{
  unsigned dt;
  if (!stop) {
  #if TEST_PIDCALIB
    //
    // Test-Updates nur im vorgegebenen Zeitraster ausführen. (Dazwischen
    //  müsste mit einer Auflösung < 0.1°C interpoliert werden.)
    //
    if (nexttestmilli > millis()) {   // wait for next test sample 
      return 2;
    }
    if (testI < (int)ELEMENTS(testT) - 1) {
      testI++;				// next test value
    }
    nexttestmilli += test_dt;
    for (int i = 0; i < 8; i++) {	// PIDCal->Update() builds mean of 8!
      dt = PIDCal->Update(-testT[testI]);  // time units (128 msec) expired since previous update
    }
  #else
    dt = PIDCal->Update(-ReadTemperature_xS());  // time units (128 msec) expired since previous update
  #endif
  }
  if (stop  ||  dt >= (int)(600L * 1000 / 128)) { 	// stop 600s w/ new sample
    //
    // finish if user break or more than 5 minutes since previous update:
    // 
    PID_delay = -1;			// stop calibrating
    calibState = 0;
  #if !TEST_PIDCALIB
    SetCoolingLevel(coolingLevel	// return to pre-calib cooling level
      #if COOLER_PWM_REVERSE	
	-
      #else 
	+
      #endif
          PIDCal->GetOutputStep());
  #endif
    int r = PIDCal->CalculateCoeffs();
    if (r == 0) {
      bluePrintln_P(PSTR("\nPID-Kalibrierung fehlgeschlagen"));
    }
    else {
      calibState = 4 | r;			    // Rest stückweise ausführen
    }
    return -1;
  }
  if (PIDCal->GetCount() != PID_calCnt) {
    PID_calCnt = PIDCal->GetCount();
    if (debug) {
      blueSerial.print(PID_calCnt);
      bluePrint_P(PSTR(": T="));
      bluePrintTdegC(-PIDCal->GetLastReading_x1());
      bluePrint_P(PSTR(", dt="));
      bluePrintDecimals((millis() - PIDCal->GetStartMillis()) / 100, 1);
      bluePrintln_P(PSTR("s"));
    }
    else {
      bluePrint_P(PSTR("."));	// indicate next response level reached
    }
  }
  return 2;		    // calibrating
}


//
// After PID calibration, let user decide what to do.
//
void IntervallTimer::ContinueCalib()
{
  switch (calibState & ~3) {
    case 4:
      if (calibState & 1) {
	bluePrint_P(PSTR("\nPID-Kalibrierung (PWM="));
	bluePrint(coolingLevel);
	bluePrint_P(PSTR(" -> "));
	bluePrint(coolingLevel 
	  #if COOLER_PWM_REVERSE	
	   +
	  #else 
	   -
	  #endif
	    PIDCal->GetOutputStep());
	bluePrint_P(PSTR("):\n1.) "));
	PrintPID(PIDCal->Get_T_sum());
	bluePrint_P(PSTR(" (nach I-TAE) anwenden\n2.) ... dauerhaft eintragen"));
      }
    break;
    case 8:
      if (calibState & 2) {
        bluePrint_P(PSTR("\n3.) "));
	PrintPID(PIDCal->Get_ZieglerNichols());
        bluePrint_P(PSTR(" (Ziegler–Nichols) anwenden\n4.) ... dauerhaft eintragen"));
      }
    break;
    case 12:
        bluePrint_P(PSTR("\n5.) Neu anzeigen\nAuswahl (0 um abzubrechen): "));
    break;
    case 16: {
      if (!blueSerial.available()) {	// Zeichen über BT da?
	return;				//  nein
      }
      uint8_t z = blueSerial.read() - '0';	// nur 0 - 4 akzeptieren
      if (z == 5) {
	PID_calCnt = 0;			// neu anzeigen  
	break;
      }
      if (	z > 4  
	  ||  ((z == 1 || z == 2)  &&  (calibState & 1) == 0)
	  ||  (z >= 3  &&  (calibState & 2) == 0)
	) {
        calibState -= 4;      // query again
	return;
      }
      calibState = 0;
      float* kkk = PIDCal->Get_T_sum();	  // für z=1,2
      switch (z) {
	case 0:	
	  bluePrintln_P(PSTR("\nPID-Kalibrierung abgebrochen\n"));
	  return;
	case 3:
	case 4: kkk = PIDCal->Get_ZieglerNichols(); break;
      }
      for (int i = 0; i < 3; i++) {
	KpKiKd[i] = kkk[i];
      }
      if (z == 2  ||  z == 4) {	  	  // write to EEPROM
	SavePIDtoEEPROM();
        bluePrintln_P(PSTR("\nPID-Koeffizienten im EEPROM gespeichert\n"));
      }
      else {
        bluePrintln_P(PSTR("\nPID-Kalibrierung abgeschlossen\n"));
      }
      return;
    }
    case 20: {			  // neu anzeigen:
      if (PID_calCnt < PIDCal->GetCount()) {
	blueSerial.print(PID_calCnt);
	bluePrint_P(PSTR(": T="));
	bluePrintTdegC(-PIDCal->GetReading_x1(PID_calCnt));
	bluePrint_P(PSTR(", dt="));
	bluePrintDecimals((int)(PIDCal->GetTime(PID_calCnt) * 1.28), 1);// *128 /1000 *10
	bluePrintln_P(PSTR("s"));
	PID_calCnt++;
	return;
      }
      calibState -= 16;	    // Koeffs neu anzeigen
      return;
    }
  }
  calibState += 4;
}


//
// Load PID coeffs from EEPROM or basic value set.
//
// Returns truee if loaded from EEPROM
//
bool IntervallTimer::ResetPID(bool fromEEPROM)
{
  if (fromEEPROM && LoadPIDfromEEPROM()) {
    return true;
  }
  KpKiKd[0] = PID_Kp;
  KpKiKd[1] = PID_Ki;
  KpKiKd[2] = PID_Kd;
  return false;
}


//
// Print PID coeffs
//
void IntervallTimer::PrintPID(float* kkk)
{
  bluePrint_P(PSTR("Kp = "));
  bluePrintDecimals((int)(kkk[0] * 100), 2);
  bluePrint_P(PSTR(", Ki = "));
  bluePrintDecimals((int)(kkk[1] * 1000), 3);
  bluePrint_P(PSTR(", Kd = "));
  if (fabs(kkk[2]) > 300) {
    bluePrintDecimals((int)(kkk[2] * 10), 1);
  }
  else {
    bluePrintDecimals((int)(kkk[2] * 100), 2);
  }
}

#endif // COOLER_PWM
